import json
from typing import Dict, List

COMPANY_NAME_TO_CIK = "company_name_to_cik"
TICKER_TO_CIK = "ticker_to_cik"

API_STRUCTURE_FIELD_TICKER = "ticker"
API_STRUCTURE_FIELD_TITLE = "title"
API_STRUCTURE_FIELD_CIK = "cik_str"


class CompanyTickerHandler:
    company_name_to_cik: Dict[str, int]
    ticker_to_cik: Dict[str, int]

    def __init__(self, company_tickers_json: dict):
        self.company_name_to_cik = {}
        self.ticker_to_cik = {}

        if '0' in company_tickers_json.keys():
            self.load_from_sec_api_structure(company_tickers_json)
        elif COMPANY_NAME_TO_CIK in company_tickers_json and TICKER_TO_CIK in company_tickers_json:
            self.load_from_file_structure(company_tickers_json)
        else:
            raise ValueError("Unknown format for company tickers JSON")

    def load_from_file_structure(self, company_tickers_json: dict):
        """
        Loads the json from the file structure rather than the api structure
        :param company_tickers_json: The json file to load into the handler
        """
        self.company_name_to_cik = company_tickers_json[COMPANY_NAME_TO_CIK]
        self.ticker_to_cik = company_tickers_json[TICKER_TO_CIK]

    def load_from_sec_api_structure(self, company_tickers_json: dict):
        """
        Loads the json from the api structure into the correct lookup tables
        :param company_tickers_json: The json file to load into the handler
        """
        for _, company_tick_info in company_tickers_json.items():
            cik = company_tick_info[API_STRUCTURE_FIELD_CIK]
            title = company_tick_info[API_STRUCTURE_FIELD_CIK]
            ticker = company_tick_info[API_STRUCTURE_FIELD_TICKER]
            if isinstance(cik, int) or len(cik) > 0:
                self.company_name_to_cik[title] = int(cik)
                self.ticker_to_cik[ticker] = int(cik)

    def store_company_tickers_to_file(self, filepath: str):
        """
        Store the company tickers loaded into the file to the given path
        :param filepath: The path to the file where to store the loaded tickers.
        """
        fd = open(filepath, "w+")
        ticker_dump = {
            COMPANY_NAME_TO_CIK: self.company_name_to_cik,
            TICKER_TO_CIK: self.ticker_to_cik
        }
        fd.write(json.dumps(ticker_dump))

    def get_cik_to_ticker(self, ticker_name: str) -> int:
        """
        Returns the CIK to the ticker name
        :param ticker_name: The ticker name for which we want to know the CIK number
        :return: The CIK retrieved.
        """
        return self.ticker_to_cik[ticker_name]

    def get_cik_to_company_name(self, company_name: str) -> List[int]:
        """
        Returns the CIK numbers all names matching the company name query
        :param company_name: The company name query for which we want to know the CIK numbers.
        :return: The CIK retrieved.
        """
        raise NotImplementedError("Retrieving the cik to the company name is currently not supported")



from pprint import pformat, pprint
from typing import List, Dict, Optional

from ..data_representation.SECBaseMeasure import SECBaseMeasure
from ..general.SECConstants import DEI_API_FIELD, US_MINUS_GAAP_API_FIELD, SRT_FIELD, \
    IFRS_MINUS_FULL_API_FIELD, INVEST_FIELD, DOCUMENT_AND_ENTITY_INFORMATION_FIELD, US_GAAP_FIELD, COMPANY_NAME_FIELD, \
    TICKER_NAME_FIELD, CIK_FIELD, IFRS_FULL_FIELD


class SECFilingInfo:
    document_and_entity_information: List[SECBaseMeasure]
    us_gaap: List[SECBaseMeasure]
    srt: List[SECBaseMeasure]
    invest: List[SECBaseMeasure]
    ifrs_full: List[SECBaseMeasure]
    _base_measure_map: Dict[str, SECBaseMeasure]

    def __repr__(self):
        return pformat(self.__dict__)

    def _process_base_measures(self, data: dict, key: str, delete_key: bool = True):
        """
        Retrieves a base measure from the data dictionary based on the given key and creates the corresponding objects.
        Also fills the lookup map for base measures
        :param data: Data dictionary containing information about the companies filings
        :param key: The key which should be extracted and converted to base measures
        :param delete_key: If true the key and value for the key are deleted from the data dictionary
        """
        if key in data:
            for base_measure_name, base_measure_entry in data[key].items():
                sec_base_measure = SECBaseMeasure(base_measure_name, **base_measure_entry)
                self.document_and_entity_information.append(sec_base_measure)
                self._base_measure_map[base_measure_name] = sec_base_measure

            if delete_key:
                data.pop(key)

    def __init__(self, data: dict):
        self.document_and_entity_information = []
        self.us_gaap = []
        self.srt = []
        self.invest = []
        self.ifrs_full = []
        self._base_measure_map = {}
        self._process_base_measures(data, DEI_API_FIELD, delete_key=True)
        self._process_base_measures(data, US_MINUS_GAAP_API_FIELD, delete_key=True)

        if SRT_FIELD in data.keys():
            self._process_base_measures(data, SRT_FIELD, delete_key=True)
        if IFRS_MINUS_FULL_API_FIELD in data.keys():
            self._process_base_measures(data, IFRS_MINUS_FULL_API_FIELD, delete_key=True)
        if INVEST_FIELD in data.keys():
            self._process_base_measures(data, INVEST_FIELD, delete_key=True)

        pprint(self._base_measure_map.keys())

        if len(data) > 0:
            print(f"Unknown keys {data.keys()} in object")

    def get_measure_info(self, name: str) -> Optional[SECBaseMeasure]:
        """
        Returns the measure info to the given name
        :param name: The name of the measure to extract
        :return: The SECBaseMeasure when found, None otherwise
        """
        return self._base_measure_map.get(name, None)

    def to_dict(self):
        return {
            DOCUMENT_AND_ENTITY_INFORMATION_FIELD: [x.to_dict() for x in self.document_and_entity_information],
            US_GAAP_FIELD: [x.to_dict() for x in self.us_gaap],
            SRT_FIELD: [x.to_dict() for x in self.srt],
            INVEST_FIELD: [x.to_dict() for x in self.invest],
            IFRS_FULL_FIELD: [x.to_dict() for x in self.ifrs_full],
        }

    def measures_to_list_of_dicts(self, company_name: str, ticker_name: str, cik: str) -> List[dict]:
        """
        Converts the SECBaseMeasure inside the SECFilingInfo into a list of SECBaseMeasures containing the company name
        and the ticker name of the company as well as the CIK number
        :param company_name: The name of the company
        :param ticker_name: The ticker name of the company
        :param cik: The cik number as a string with leading zeros
        :return: List of dicts containing the base measures.
        """
        company_info = {COMPANY_NAME_FIELD: company_name, TICKER_NAME_FIELD: ticker_name, CIK_FIELD: cik}
        base_measure = []
        base_measure.extend([{**company_info, **x.to_dict()} for x in self.document_and_entity_information])
        base_measure.extend([{**company_info, **x.to_dict()} for x in self.us_gaap])
        base_measure.extend([{**company_info, **x.to_dict()} for x in self.srt])
        base_measure.extend([{**company_info, **x.to_dict()} for x in self.invest])
        base_measure.extend([{**company_info, **x.to_dict()} for x in self.ifrs_full])
        return base_measure



from datetime import datetime
from typing import Optional

from ..general.Constants import DATE_FORMAT


class SECUnits:
    unit: str
    start: Optional[datetime]
    end: datetime
    val: int
    accn: str
    year: int
    quarter: str
    form_type: str
    date_filed: datetime
    frame: str

    def __init__(self, end: str, val: int, accn: str, fy: int, fp: str, form: str, filed: str,
                 unit: str, frame: str = "", start: Optional[str] = None):
        self.end = datetime.strptime(end, DATE_FORMAT)
        self.date_filed = datetime.strptime(filed, DATE_FORMAT)
        self.val = val
        self.accn = accn
        self.year = fy
        self.quarter = fp
        self.form_type = form
        self.unit = unit
        self.frame = frame
        self.start = datetime.strptime(start, DATE_FORMAT) if start is not None else start

    def to_dict(self):
        return self.__dict__



import json
import os
import time
from time import perf_counter_ns
from typing import Union, Any, List, Dict

import requests

from ..data_loading.CompanyTickerHandler import CompanyTickerHandler
from ..data_representation.SECCompanyInfo import SECCompanyInfo

CACHE_DIRECTORY = "SECAPICache"
COMPANY_TICKERS_FILE_NAME = "company_tickers.json"


class SECAPILoader:
    company_tickers_url: str
    data_api_url_format: str
    company_facts_url_format: str
    request_headers: dict
    ticker_handler: CompanyTickerHandler
    _last_access_time: int

    def __init__(self,
                 your_company_name: str,
                 your_company_email: str,
                 company_tickers_url: str = r"https://www.sec.gov/files/company_tickers.json",
                 data_api_url_format: str = "https://data.sec.gov/submissions/CIK{:010d}.json",
                 company_facts_url_format: str = "https://data.sec.gov/api/xbrl/companyfacts/CIK{:010d}.json",
                 override_cached_company_tickers: bool = False):
        """
        Initializes the api loader responsible for retrieving the data over the SEC-Api. To fit the fair use rules
        of the SEC you need to provide your name and email address as user agent
        :param company_tickers_url: The company tickers url, which contains a JSON-file with ticker name,
            CIK number and the company name. This JSON is responsible for mapping ticker names or company names to
            CIK numbers for fetching
        :param data_api_url_format: The format string representing the data url, which should be used for retrieval
            of data over the SEC api
        :param override_cached_company_tickers: Overrides company tickers stored locally, by reloading them on init
            of the api loader with the json given by company_tickers_url.
        """
        self.request_headers = {
            'User-Agent': your_company_name,
            'From': your_company_email,
            "Accept-Encoding": "gzip, deflate",
            "Host": "data.sec.gov"
        }
        self._last_access_time = 0
        self.company_tickers_url = company_tickers_url
        self.data_api_url_format = data_api_url_format
        self.company_facts_url_format = company_facts_url_format

        if not os.path.exists(CACHE_DIRECTORY):
            os.mkdir(CACHE_DIRECTORY)

        if override_cached_company_tickers or (COMPANY_TICKERS_FILE_NAME not in os.listdir(CACHE_DIRECTORY)):
            # Load the tickers from the url
            response = requests.post(self.company_tickers_url, headers=self.request_headers)

            if not response.ok:
                raise Exception("Not able to load the company tickers over the SEC-Api. "
                                f"Reason: {response.reason}, Status Code: {response.status_code}\n"
                                f"Response Text: \n{response.text}")

            self.ticker_handler = CompanyTickerHandler(json.loads(response.text))
            self.ticker_handler.store_company_tickers_to_file(f"{CACHE_DIRECTORY}/{COMPANY_TICKERS_FILE_NAME}")
        else:
            tickers_json = json.load(open(f"{CACHE_DIRECTORY}/{COMPANY_TICKERS_FILE_NAME}", "r"))
            self.ticker_handler = CompanyTickerHandler(tickers_json)

    def _handle_fair_use_rate_limit(self, limit_per_second: int = 10):
        current_time = perf_counter_ns()
        # Calc delta in seconds
        delta = (current_time - self._last_access_time) / 10 ** 9
        time_until_next_call = 1 / limit_per_second

        if delta < time_until_next_call:
            time.sleep(time_until_next_call - delta)
        self._last_access_time = current_time

    def _retrieve_from_api_with_ticker(self, ticker_name: str, format_url: str) -> dict:
        """
        Retrieves the data from the given format url by fetching via ticker name and CIK
        :param ticker_name: The ticker name to retrieve
        :param format_url: The format url to use for fetching
        :return: The data as dict.
        """

        cik = self.ticker_handler.get_cik_to_ticker(ticker_name)
        data_url = format_url.format(cik)
        self._handle_fair_use_rate_limit()
        response = requests.get(data_url, headers=self.request_headers)

        if response.ok:
            data = response.json()
        else:
            raise ConnectionError(f"Retrieved error code {response.status_code} with message {response.text} "
                                  f"while retrieving data for ticker {ticker_name} resulting in "
                                  f"api url: {data_url}")

        return data

    def get_data_from_ticker(self, ticker_name: str, as_dict: bool = False) -> Union[Any, dict]:
        """
        TODO define a class holding the filings
        Get the data either as raw dict or as classes
        :param ticker_name: The name of the ticker to retrieve
        :param as_dict: If true the data is returned as dict instead of SECCompanyInfo object.
        :return: The company info retrieved.
        """
        return self._retrieve_from_api_with_ticker(ticker_name, self.data_api_url_format)

    def get_company_facts_by_ticker(self, ticker_name: str, as_dict: bool = False):
        """
        Get the data either as raw dict or as classes
        :param ticker_name: The name of the ticker to retrieve
        :param as_dict: If true the data is returned as dict instead of SECCompanyInfo object.
        :return: The company info retrieved.
        """

        data = self._retrieve_from_api_with_ticker(ticker_name, self.company_facts_url_format)
        if as_dict:
            return data

        return SECCompanyInfo(**data)

    def get_company_facts_by_ticker_list(self, ticker_list: List[str],
                                         as_dict: bool = False) -> Dict[str, Union[SECCompanyInfo, dict]]:
        """
        Retrieves company facts by a list of tickers passed
        :param ticker_list: The list of tickers to retrieve
        :param as_dict: If true the data is returned as dictionary, otherwise each company info is converted
            to a SECCompanyInfo object
        :return: A dictionary with the ticker names as key and the SECCompanyInfo or dictionaries of data as values
        """
        result_dict = {}

        for ticker_name in ticker_list:
            result_dict[ticker_name] = self.get_company_facts_by_ticker(ticker_name, as_dict)

        return result_dict



from typing import Optional, Union, List

from ..data_representation.SECBaseMeasure import SECBaseMeasure
from ..data_representation.SECFilingInfo import SECFilingInfo


class SECCompanyInfo:
    cik: str
    company_name: str
    ticker_name: str
    facts: SECFilingInfo

    def __init__(self, cik: Union[int, str], entityName: str, facts: dict):
        self.cik = str(cik).zfill(10)
        self.company_name = entityName
        self.facts = SECFilingInfo(facts)
        self.ticker_name = ""

    def get_measure_info(self, name: str) -> Optional[SECBaseMeasure]:
        """
        Returns the measure info to the given name
        :param name: The name of the measure to extract
        :return: The SECBaseMeasure when found, None otherwise
        """
        return self.facts.get_measure_info(name)

    def set_ticker_name(self, ticker_name: str):
        self.ticker_name = ticker_name

    def to_dict(self):
        my_dict = self.__dict__
        my_dict["facts"] = self.facts.to_dict()
        return my_dict

    def measures_to_list_of_dicts(self) -> List[dict]:
        """
        Converts the SECBaseMeasure inside the SECFilingInfo into a list of SECBaseMeasures containing the company name
        and the ticker name of the company as well as the CIK number.
        """
        return self.facts.measures_to_list_of_dicts(self.company_name, self.ticker_name, self.cik)

    def get_document_entity_information_names(self) -> List[str]:
        """
        Retrieves the names from the document entity information of the SECFilingInfo of the company
        representing certain information like the NumberOfSharesOutstanding.
        :return: The names of measures, which are available for the company.
        """

        return [measure.name for measure in self.facts.document_and_entity_information]

    def get_document_entity_information_labels(self) -> List[str]:
        """
        Retrieves the labels from the document entity information of the SECFilingInfo of the company
        representing certain information like the NumberOfSharesOutstanding.
        :return: The labels of measures, which are available for the company.
        """

        return [measure.label for measure in self.facts.document_and_entity_information]



"""File containing constants, which are used for reading the data from the api JSON files"""
TICKER_NAME_FIELD = "ticker_name"
COMPANY_NAME_FIELD = "company_name"
CIK_FIELD = "cik"
DOCUMENT_AND_ENTITY_INFORMATION_FIELD = "document_and_entity_information"
IFRS_FULL_FIELD = "ifrs_full"
US_GAAP_FIELD = "us_gaap"
INVEST_FIELD = "invest"
IFRS_MINUS_FULL_API_FIELD = "ifrs-full"
SRT_FIELD = "srt"
US_MINUS_GAAP_API_FIELD = "us-gaap"
DEI_API_FIELD = "dei"



from collections import defaultdict
from pprint import pformat
from typing import List, Dict, Optional

from ..data_representation.SECUnits import SECUnits


class SECBaseMeasure:
    name: str
    label: str
    description: str
    units: List[SECUnits]
    _form_type_to_unit: Dict[str, List[SECUnits]]

    def __repr__(self):
        return pformat(self.__dict__)

    def __init__(self, name: str, label: str, description: str, units: Dict[str, List[Dict]]):
        self.name = name
        self.label = label
        self.description = description
        self.units = []
        self._form_type_to_unit = defaultdict(list)
        self._parse_sec_units(units)

    def _parse_sec_units(self, units: Dict[str, List[Dict]]):
        """Parses the SECUnits into a flat list of values"""
        for unit, list_of_sec_unit_info in units.items():
            for sec_unit_info in list_of_sec_unit_info:
                self.units.append(SECUnits(unit=unit, **sec_unit_info))
                self._form_type_to_unit[self.units[-1].form_type].append(self.units[-1])

    def get_by_form_type(self, form_type: str) -> Optional[List[SECUnits]]:
        """
        Returns the list of sec units for the given form type for the base measure
        :param form_type: The form type to which we want to retrieve the data e.g. 10-K for yearly report or
            10-Q for retrieving the quarterly reports.
        :return: The list of sec units when found, None otherwise
        """
        return self._form_type_to_unit.get(form_type, None)

    def to_dict(self):
        return {
            "name": self.name,
            "label": self.label,
            "description": self.description,
            "units": [unit.to_dict() for unit in self.units]
        }



"""Class maintaining the plotting functionality for SECUnit entries."""
from enum import Enum
from typing import List, Tuple

import matplotlib.pyplot as plt
import seaborn as sns

from ..data_representation.SECUnits import SECUnits


class NumberFormat(Enum):
    """Enum for defining the number format."""
    PLAIN = 0
    THOUSANDS = 1
    MILLIONS = 2
    BILLIONS = 3
    TRILLIONS = 4


class SECUnitPlotter:
    _palette = None

    def __init__(self, color_palette=None):
        sns.set_theme()

        if color_palette is not None:
            self._palette = sns.color_palette("coolwarm", as_cmap=True)
        else:
            self._palette = color_palette

    def _apply_number_format(self, list_of_values: List[float], number_format: NumberFormat) -> Tuple[List[float], str]:
        return [round(val / 10 ** (3 * number_format.value), 3) for val in list_of_values], number_format.name

    def _determine_auto_format(self, list_of_values: List[float]) -> NumberFormat:
        smallest_val = min(list_of_values)

        if (smallest_val / 10 ** (3 * NumberFormat.TRILLIONS.value)) > 1:
            return NumberFormat.TRILLIONS
        elif (smallest_val / 10 ** (3 * NumberFormat.BILLIONS.value)) > 1:
            return NumberFormat.BILLIONS
        elif (smallest_val / 10 ** (3 * NumberFormat.MILLIONS.value)) > 1:
            return NumberFormat.MILLIONS
        elif (smallest_val / 10 ** (3 * NumberFormat.THOUSANDS.value)) > 1:
            return NumberFormat.THOUSANDS
        elif (smallest_val / 10 ** (3 * NumberFormat.PLAIN.value)) > 1:
            return NumberFormat.PLAIN

    def plot_sec_units_base(self, sec_units: List[SECUnits], kind: str = "bar",
                            number_format: NumberFormat = None,
                            auto_number_format: bool = True, title: str = None):
        """
        Base functionality for plotting a list of SECUnit entries
        :param title: The title of the plot
        :param auto_number_format: In case no number format is passed and auto number format should be used,
            the best format suitable format is chosen for displaying
        :param number_format: The number format to use for displaying the y-values
        :param kind: The kind of plot either bar or line chart
        :param sec_units: The list of SECUnits to plot
        """
        sec_units_ordered = sorted(sec_units, key=lambda x: x.end)

        if len(sec_units_ordered) == 0:
            raise AttributeError("Empty List passed for plotting the SECUnits!")

        x_values = [unit.end.strftime("%Y-%m-%d") for unit in sec_units_ordered]
        y_values = [unit.val for unit in sec_units_ordered]

        if number_format is None and auto_number_format:
            number_format = self._determine_auto_format(y_values)

        if number_format is not None and number_format.value != 0:
            y_values, format_label = self._apply_number_format(y_values, number_format)
            plt.ylabel(f"({sec_units_ordered[-1].unit}) in {format_label}")

        if kind == "bar":
            ax = sns.barplot(x=x_values, y=y_values, palette=self._palette, color='b')
        elif kind == "line":
            ax = sns.lineplot(x=x_values, y=y_values, palette=self._palette)
        else:
            raise AttributeError("Passed unsupported type for kind of plot. Parameter 'kind' can only take values "
                                 "'line' or 'bar'")

        plt.xticks(rotation=45)
        plt.ticklabel_format(style='plain', axis='y')
        if title is not None:
            plt.title(title)
        plt.show()



import json
import os
import traceback
from multiprocessing import Pool, Lock
from pprint import pprint

from pymongo import MongoClient
from pymongo.collection import Collection

from ..data_representation.SECCompanyInfo import SECCompanyInfo

lock: Lock


def init_multiprocessing(multi_proc_lock: Lock):
    global lock
    lock = multi_proc_lock


def load_data(docs, path_to_filings, tickers_cik, tickers_name, start_index, end_index):
    global lock
    data_retrieved = []
    for i in range(start_index, end_index):
        print(f"\rCurrent index {i}", end="")
        company_data = json.load(open(f"{path_to_filings}/{docs[i]}"))
        if len(company_data) == 0:
            return

        try:
            company_info = SECCompanyInfo(**company_data)
            # print(f"Entity Name: {company_info.entity_name}")
            if company_info.cik in tickers_cik:
                company_info.set_ticker_name(tickers_cik[company_info.cik])
            elif company_info.company_name in tickers_name:
                company_info.set_ticker_name(tickers_name[company_info.company_name])
            else:
                pass
                # print(f"Unknown CIK {company_info.cik}")

            lock.acquire()
            data_retrieved.extend(company_info.measures_to_list_of_dicts())
            lock.release()

            # print(company_info.ticker_name)
        except Exception as e:
            lock.release()
            pprint(traceback.format_exc())

    # print(data_retrieved)
    try:
        conn = MongoClient()
        db = conn["SECData"]
        coll = db["SECInfo"]
        coll.insert_many(data_retrieved)
    except Exception:
        pprint(traceback.format_exc())
        exit()


class SECFileDataLoader:

    def __init__(self, path_to_filings: str, path_to_tickers: str, n: int, j: int = 6,
                 ip: str = "127.0.0.1", port: int = 27017, number_of_entries_per_process: int = 50):
        """
        Initializes the data loader and loads the data
        :param path_to_filings: The path to the sec filings
        :param path_to_tickers: The path to the ticker file for mapping
        :param n: The number of documents to read if -1 is passed all documents in the path_to_filings
                  directory are read
        :param j: The number of workers for reading the files.
        """
        conn = MongoClient(ip, port)
        db = conn["SECData"]
        coll = db["SECInfo"]
        coll.drop()
        coll.insert_one({"DUMMY": "DUMMY"})
        l = Lock()
        tickers = json.load(open(path_to_tickers))
        tickers_cik = {str(t["cik_str"]).zfill(10): t["ticker"] for key, t in tickers.items()}
        tickers_name = {t["title"]: t["ticker"] for key, t in tickers.items()}
        docs = os.listdir(path_to_filings)
        self.data = []
        if n == -1:
            n = len(docs)

        results = []
        with Pool(processes=j, initializer=init_multiprocessing, initargs=(l,)) as pool:
            for i in range(0, n, number_of_entries_per_process):
                if i < len(docs):
                    res = pool.apply_async(load_data,
                                           args=(docs, path_to_filings, tickers_cik, tickers_name, i,
                                                 i + number_of_entries_per_process,))
                    results.append(res)
                print(f"\rProcessed res_count {i}/{n}!", end="")

            pool.close()
            pool.join()
        print(f"Finished!")

    def mongodb_data_to_from_secinfo_to_secunits(self, coll: Collection):
        """
        Splits the SECInfo entries, which contains the SECBaseMeasures into its SECUnits. Descriptions and
        the units are placed in the root document.
        """
        coll.aggregate([
            {
                '$unwind': {
                    'path': '$units'
                }
            }, {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [
                            '$$ROOT', '$units'
                        ]
                    }
                }
            }, {
                '$project': {
                    'units': 0,
                    '_id': 0
                }
            }, {
                '$out': 'SECUnits'
            }
        ])
